;
; Ordering Routines DLL (currently only the comb order)
; 2002-03-17
; Dwayne Robinson
;
; Written by request of and for Brian, to compensate for Visual Basic's
; ability to make a fast new computer seem like it's yesterday's model.
; Only after all other possible optimizations have been made should something
; be ported to assembler. Either that, or when you have TONs of data to order.
;
;
; Current progress:
;
;   Complete as far as I'm concerned...
;
;
; Fixed:
;
; Seemingly infinite loop:
;   It wasn't actually an infinite loop after all. I was just too impatient
;   to let it finish. It only happens when the number of items is HUGE,
;   like 1 million or so. It occurs because a rare straggler somehow cheats
;   the first pass and causes the normally fast comb ordering to degrade
;   into the much slower bubble ordering. So now when there are one or two
;   such miscreants, I simply take another pass to put them in their place.
;
; Visual Basic singles not working:
;   FPU flag bits are in a different order than the CPU flags, so moving
;   them to ax (fnstsw) and using JG just doesn't work. Instead, you must
;   TEST the bits and use JNZ/JZ. Also, I was using FCOM instead of FCOMP,
;   so after comparing only eight entries, VB would report the error
;   message "expression too complex". Huh? The TRUE error was actually FPU
;   stack overflow.
;
;
;
; Parameters:
;
; array ptr
;   Pointer to first element of array. It only accepts 32bit data, meaning
;   the ordering routine can't work with byte or short arrays; but this
;   limitation should be acceptable considering longs, singles, and string
;   ptrs are generally types that need to be ordered anyway. The exact type
;   depends on what the comparison callback compares.
;
; array size
;   Number of elements, not byte size. Pass the actual number of elements
;   in the array, not UBOUND (which returns one less than the actual size).
;
; comparison callback
;   For versatility, the ordering routines do not any comparisons
;   themselves. Instead, they use the reply of a called function to
;   determine whether or not to swap values. The callback is provided two
;   parameters to compare, which it must then return true if they should be
;   swapped, false if they need not be.
;
;   Some common callbacks are provided local to the DLL, so that you need
;   not define your own in a high level language (which would counteract
;   the advantage of this being written in assembler). These internal
;   callbacks are faster because they can simply use the registers
;   (esi & edi) rather than setting up the whole stack frame just for the
;   sake of a mere two variables.
;
; escape callback
;   Since the ordering routine does NOT return to your own code until done,
;   this callback allows to check whether the user wants to abort the
;   ordering and gives an opportunity to update a progress meter. To abort,
;   it should return TRUE, otherwise the order continues. This routine is
;   only called only once each complete pass. It has no parameters.
;
;   Pass NULL if you do not want an escape callback.
;
;
;
; Limitations:
;
; in memory order
;   The entire array must be located in memory. Actually, it is possible
;   if the comparison callback supported it, to also order disk records.
;   For example, if each entry was actually a file pointer. If you want
;   to do that though, don't even bother using this library because any
;   speed increases would be negligible. Also ensure you actually have
;   enough memory or else be prepared for major thrashing (ie. 10 million
;   longs on a 32MB machine).
;
; 32bit array data
;   It does not matter what type of 32 bit data it is, just as long as it
;   each entry in the array is 32bit (longs, singles, string ptrs). The
;   comparison callback can then take care of the type specifics. This
;   limitation restricts the algorithm from ordering byte arrays, but
;   ordering byte arrays really isn't one of the top requests.
;
;
;
; Visual Basic function definitions:
;
; Declare Sub CombOrder Lib "asmorder" (Array() As Long, _
;                                       ByVal ArraySize As Long, _
;                                       ByVal CmpCallback As Long, _
;                                       ByVal EscCallback As Long)
;
; Public Function CmpCallbackString (ByRef Parm1 As String, _
;                                     ByRef Parm2 As String)
;
; Public Function CmpCallbackLong (ByVal Parm1 As Long, _
;                                  ByVal Parm2 As Long)
;
; Note that because VB is bad at finding DLLs, you must either:
; -Start it from the project (NOT open VB and then simply open the project) so
;  that the current directory points correctly. In other words, double click
;  on the VBP file.
; -Include the full path in the Lib statement.
; -Compile the program and run that.
; -Copy the DLL to WINDOWS\SYSTEM (bad choice).
;
;
;
; Calling from Visual Basic:
;
; Const CmpLongAsc  As Long = 0    <- 32bit signed integers
; Const CmpLongDesc As Long = 1
; Const CmpSngAsc   As Long = 2    <- floating point singles
; Const CmpSngDesc  As Long = 3
; Const CmpStrAscA  As Long = 4    <- for C strings (ASCII)
; Const CmpStrDescA As Long = 5
; Const CmpStrAscW  As Long = 6    <- for Unicode C strings (Widebyte)
; Const CmpStrDescW As Long = 7
; Const CmpStrAscB  As Long = 8    <- for Visual Basic 4+ strings (BSTRs)
; Const CmpStrDescB As Long = 9
;
; Const TestArraySize As Long = 500
; Dim TestArray(TestArraySize-1) As Long
;
; CombOrder TestArray(), TestArraySize, CmpLongAsc, 0
; CombOrder TestArray(), TestArraySize, AddressOf(MyOwnCmp), 0
;
;
;

[section data data]
[section code code]
[section bss bss]

global DllMain
global CombOrder

;%define debug
;%define profiling
;%define UsePeImages
%include "mywininc.asm"

MaxStrLength equ 1024

;
section data
ProgTitle:  db "AsmOrder",0
            db "Dwayne Robinson",0
            db "Peekin",0
            db "FDwR@hotmail.com",0
            db "2002-03-20",0

align 4,db 0
CmpFunc:    dd CmpLongAsc       ;callback to perform comparison
EscFunc:    dd NoEscCallback

CmpTblEntries equ 10
CmpTbl:     dd CmpLongAsc
            dd CmpLongDesc
            dd CmpSngAsc
            dd CmpSngDesc
            dd CmpStrAscA
            dd CmpStrDescA
            dd CmpStrAscW
            dd CmpStrDescW
            dd CmpStrAscB
            dd CmpStrDescB

%ifdef profiling
Swaps:      dd 0                ;number of swaps
Comps:      dd 0                ;number of comparisons
%endif
Swapped:    dd 0                ;flag that values have been swapped

section code
;
; ORDERING ROUTINES
;
; Doesn't do much, but what it does do is important for any program that
; calls this DLL. It will start fine the first time, but any attempts to
; start the same program later causes Windows to claim it is an invalid EXE
; format (which is totally ludicrous!?). So when the linker gives a warning
; about no entry point, take it seriously.

DllMain:
    ;api MessageBox, 0, Author, ProgTitle, 0
    mov eax,TRUE
    ret 12

;
; (dword array ptr,
;  dword array size,
;  dword comparison callback,
;  dword escape callback)
; (nothing)
CombOrder:
    debugmsg "dll stack=%x",esp

    paramsrel 16, .ArrayPtr, .ArraySize, .CmpFunc, .EscFunc
    push ebx,esi,edi            ;save these so VB doesn't get mad

    debugmsg "ArrayPtr=%x ArraySize=%d",[esp+.ArrayPtr+4],[esp+.ArraySize]

    fninit                      ;for the floating point values

    mov eax,[esp+.CmpFunc]
    cmp eax,CmpTblEntries
    jae .ValidCmp
    mov eax,[CmpTbl+eax*4]      ;get local comparison callback from table
.ValidCmp:
    mov [CmpFunc],eax
    debugmsg "cmp func=%x",eax

    mov edx,[esp+.EscFunc]
    test edx,edx
    jnz .ValidEsc
    mov edx,NoEscCallback
.ValidEsc:
    debugmsg "esc func=%x",edx
    mov [EscFunc],edx

   %ifdef profiling
    xor eax,eax
    mov [Comps],eax             ;zero both the comparison and swap count
    mov [Swaps],eax
   %endif

    mov ebx,[esp+.ArraySize]    ;get initial gap entire array
.NextGap:
    ; calculate gap (x / 1.333_ == x * 3 / 4)
    lea ebx,[ebx+ebx*2]         ;*3
    shr ebx,2                   ;/4
    jnz .ValidGap
    ;inc ebx                     ;must not have gap of zero
    mov ebx,[esp+.ArraySize]    ;get initial gap entire array
    lea ebx,[ebx+ebx*2]         ;*3
    shr ebx,2                   ;/4
  .ValidGap:
    cmp ebx,10                  ;if gap is 9 or 10, force to 11, since it
    ja .No11                    ;supposedly eliminates bubbles more quickly
    cmp ebx,9
    jb .No11
    mov ebx,11
  .No11:

    ; set element comparison pointers
    mov esi,[esp+.ArrayPtr]     ;start at front of array
    lea edi,[esi+ebx*4]         ;element to compare with = base + gap
    ;debugmsg "array ptrs=%x %x",esi,edi

    ; determine count of elements to compare
    mov ecx,[esp+.ArraySize]
    mov dword [Swapped],FALSE   ;initially assume everything is in order
    sub ecx,ebx                 ;array size - gap
    jbe .InvalidCount
    push ecx
    debugmsg "count=%d gap=%d",ecx,ebx
.NextComp:

    push esi
    push edi
    call [CmpFunc]              ;determine if values should be swapped
    test eax,eax
    jz .NoSwap                  ;no swap since callback returned false

    cmp ebx,1
    ja .NoShowSwap
    debugmsg "swapping %d with %d (%X,%X)",[esi],[edi],esi,edi
.NoShowSwap:
    mov eax,[esi]
    xchg [edi],eax              ;swap values/ptrs
    mov dword [Swapped],TRUE    ;a swap occurred (so continue loop)
    mov [esi],eax
   %ifdef profiling
    inc dword [Swaps]
   %endif

.NoSwap:
   %ifdef profiling
    inc dword [Comps]
   %endif

    add esi,byte 4              ;next source element
    add edi,byte 4              ;next comparison element
    dec dword [esp]             ;one less array element
    jg .NextComp                ;loop while counter > 0
    pop ecx                     ;discard counter

.InvalidCount:
    call [EscFunc]              ;check for user abort
    test eax,eax
    jnz .Abort                  ;abort function returned true (nonzero)

    cmp ebx,1
    ja near .NextGap            ;keeping looping while gap > 1
    cmp dword [Swapped],TRUE
    je near .NextGap            ;keep looping while any values were swapped

.Abort:
    pop ebx,esi,edi             ;restore these so VB doesn't get mad and crash
    debugmsg "dll stack=%x",esp
   %ifdef profiling
    debugmsg "comps=%d swaps=%d",[Comps],[Swaps]
   %endif
    ret 16

;
; Since assembly isn't the easiest to read, here are two other examples in
; more familiar languages.

%if 0
DIM I AS LONG, J AS LONG, Temp AS LONG
CONST Shrink = 1.333333 '1.3
DIM Gap AS LONG
DIM Swapped AS INTEGER
Gap = TestSize
DO
    Gap = INT(Gap / Shrink)
    IF Gap <= 1 THEN Gap = 1
    Swapped = False
    FOR J = 0 TO TestSize - Gap - 1
        I = J + Gap
        IF Test(J) > Test(I) THEN
            Temp = Test(J)
            Test(J) = Test(I)
            Test(I) = Temp
            Swapped = True
        END IF
    NEXT J
LOOP WHILE Swapped OR Gap > 1

/**
 * A comb sort 11 sort demonstration algorithm
 * Idea from Byte April 1991 by Richard Box and Stephen Lacey
 * Michele Estebon (mestebon@vt.edu) brought this to my attention.
 * 
 * SortAlgorithm.java, Thu Oct 27 10:32:35 1994
 *
 * @author Jason Harrison@cs.ubc.ca
 * @version 	1.0, 20 Nov 1996
 *
 */
class CombSort11Algorithm extends SortAlgorithm {
    final float SHRINKFACTOR = (float)1.3;

    void sort(int a[]) throws Exception {
        boolean flipped = false;
        int gap, top;
        int i, j;

        gap = a.length;
        do {
            gap = (int) ((float) gap / SHRINKFACTOR);
            switch (gap) {
            case 0: /* the smallest gap is 1 - bubble sort */
                gap = 1;
                break;
            case 9: /* this is what makes this Combsort11 */
            case 10: 
                gap = 11;
                break;
            default: break;
            }
            flipped = false;
            top = a.length - gap;
            for (i = 0; i < top; i++) {
                if (stopRequested) {
                    return;
                }

                j = i + gap;
                if (a[i] > a[j]) {
                    int T = a[i];
                    a[i] = a[j];
                    a[j] = T;
                    flipped = true;
                }
                pause(i,j);
            }
        } while (flipped || (gap > 1));
        /* like the bubble and shell sorts we check for a clean pass */
    }
}
%endif

%ifdef Unfinished
;
BubbleOrder:
    ret

;
ShellOrder:
    ret

;
QuickOrder:
    ret
%endif

;
; DEFAULT INTERNAL COMPARISON CALLBACKS
;
; (dword source, dest, esi, edi)
; (swap TRUE/FALSE)
CmpLongAsc:
    ;debugmsg "comparing %d with %d (%X,%X)",[esi],[edi],esi,edi
    mov edx,[esi]               ;get value from base
    cmp edx,[edi]               ;compare with second value
    jg .Swap
    xor eax,eax
    ret 8
.Swap:
    mov eax,TRUE
    ret 8

CmpLongDesc:
    ;debugmsg "comparing %d with %d (%X,%X)",[esi],[edi],esi,edi
    mov edx,[esi]               ;get value from base
    cmp edx,[edi]               ;compare with second value
    jl CmpLongAsc.Swap
    xor eax,eax
    ret 8

CmpSngAsc:
    debugmsg "comparing float %f with %f (%X,%X)",[esi],[edi],esi,edi
    fld dword [esi]             ;load first value into FPU
    fcomp dword [edi]           ;compare with second value
    fnstsw ax                   ;get FPU flags
    test ah,45h                 ;test FPU flags
    jz .Swap
    xor eax,eax
    ret 8
.Swap:
    mov eax,TRUE
    ret 8

CmpSngDesc:
    debugmsg "comparing float %f with %f (%X,%X)",[esi],[edi],esi,edi
    fld dword [esi]             ;load first value into FPU
    fcomp dword [edi]           ;compare with second value
    fnstsw ax                   ;get FPU flags
    test ah,01h                 ;test FPU flags
    jnz CmpSngAsc.Swap
    xor eax,eax
    ret 8

CmpStrAscA:
    ;debugmsg "comparing %s with %s (%X,%X)",[esi],[edi],esi,edi
    cld                         ;shouldn't be necessary, but in case
    mov esi,[esi]               ;get ptr to first string
    mov edi,[edi]               ;and second string
    mov ecx,MaxStrLength        ;max length
    repe cmpsb
    pop edx                     ;pop return address
    pop edi                     ;restore ptrs for caller
    pop esi
    ja .Swap
    xor eax,eax                 ;first <= second
.Ret:
    jmp edx                     ;return to caller
.Swap:
    mov eax,TRUE                ;first > second, so return non-zero
    jmp edx                     ;return to caller

CmpStrDescA:
    ;debugmsg "comparing %s with %s (%X,%X)",[esi],[edi],esi,edi
    cld                         ;shouldn't be necessary, but in case
    mov esi,[esi]               ;get ptr to first string
    mov edi,[edi]               ;and second string
    mov ecx,MaxStrLength
    repe cmpsb
    pop edx                     ;pop return address
    pop edi                     ;restore ptrs for caller
    pop esi
    jb CmpStrAscA.Swap
    xor eax,eax                 ;first <= second
.Ret:
    jmp edx                     ;return to caller

CmpStrAscW:
    ;debugmsg "comparing %s with %s (%X,%X)",[esi],[edi],esi,edi
    cld                         ;shouldn't be necessary, but in case
    mov esi,[esi]               ;get ptr to first string
    mov edi,[edi]               ;and second string
    mov ecx,MaxStrLength        ;max length
    repe cmpsw
    pop edx                     ;pop return address
    pop edi                     ;restore ptrs for caller
    pop esi
    ja .Swap
    xor eax,eax                 ;first <= second
.Ret:
    jmp edx                     ;return to caller
.Swap:
    mov eax,TRUE                ;first > second, so return non-zero
    jmp edx                     ;return to caller

CmpStrDescW:
    ;debugmsg "comparing %s with %s (%X,%X)",[esi],[edi],esi,edi
    cld                         ;shouldn't be necessary, but in case
    mov esi,[esi]               ;get ptr to first string
    mov edi,[edi]               ;and second string
    mov ecx,MaxStrLength
    repe cmpsw
    pop edx                     ;pop return address
    pop edi                     ;restore ptrs for caller
    pop esi
    jb CmpStrAscW.Swap
    xor eax,eax                 ;first <= second
.Ret:
    jmp edx                     ;return to caller

CmpStrAscB:
    ;debugmsg "comparing %s with %s (%X,%X)",[esi],[edi],esi,edi
    cld                         ;shouldn't be necessary, but in case
    mov esi,[esi]               ;get ptr to first string
    mov edi,[edi]               ;and second string
    mov ecx,[esi-4]             ;get length of first
    mov edx,[edi-4]             ;and second string
    xor eax,eax                 ;default is no swap
    cmp ecx,edx
    jb .Shorter
    mov eax,TRUE                ;in case strings match, pick longer one
    mov ecx,edx                 ;set length to shorter one
.Shorter:
    repe cmpsw
    pop edx                     ;pop return address
    pop edi                     ;restore ptrs for caller
    pop esi
    ja .Swap
    je .Ret                     ;strings equal so compare lengths
    xor eax,eax                 ;first < second
.Ret:
    jmp edx                     ;return to caller
.Swap:
    inc eax                     ;first > second, so return non-zero
    jmp edx                     ;return to caller

CmpStrDescB:
    cld                         ;shouldn't be necessary, but in case
    mov esi,[esi]               ;get ptr to first string
    mov edi,[edi]               ;and second string
    mov edx,[esi-4]             ;get length of first
    mov ecx,[edi-4]             ;and second string
    xor eax,eax                 ;default is no swap
    cmp ecx,edx
    jb .Shorter
    mov eax,TRUE                ;in case strings match, pick longer one
    mov ecx,edx                 ;set length to shorter one
.Shorter:
    repe cmpsw
    pop edx                     ;pop return address
    pop edi                     ;restore ptrs for caller
    pop esi
    jb .Swap
    je .Ret                     ;strings equal so compare lengths
    xor eax,eax                 ;first < second
.Ret:
    jmp edx                     ;return to caller
.Swap:
    inc eax                     ;first > second, so return non-zero
    jmp edx                     ;return to caller

;
; Simply returns false. This stub is used when the calling program passes a
; NULL escape callback to the ordering routine.
NoEscCallback:
    xor eax,eax
    ret


;
; This doesn't work :(
%if 0
section resource

; IMAGE_RESOURCE_DIRECTORY
dd 0    ;Characteristics
dd 0    ;TimeDateStamp
dw 0    ;MajorVersion
dw 0    ;MinorVersion
dw 0    ;NumberOfNamedEntries
dw 1    ;NumberOfIdEntries

; IMAGE_RESOURCE_DIRECTORY_ENTRY
dd RT_VERSION   ;Name
dd VersionEntry ;OffsetToData

VersionEntry:
; IMAGE_RESOURCE_DATA_ENTRY
dd VersionData                  ;OffsetToData
dd EndVersionData-VersionData   ;Size
dd 04090000h                    ;CodePage
dd 0                            ;Reserved

VersionData:
; figure how to add these to the executable
dw EndVersionData-$             ;wLength (length of whole structure)
dw 0                            ;wValueLength (no VS_FIXEDFILEINFO)
dw 1                            ;bText (version contains text not binary)
db "VS_VERSION_INFO",0          ;WCHAR szKey[];
db "",0                         ;No VS_FIXEDFILEINFO Value

;Begin children
db "StringFileInfo",0
db "04090000",0
db "CompanyName",0,"PeekinSoft",0
db "FileDescription",0,"Comb Ordering Routine",0
db "FileVersion",0,".001",0
db "OriginalFilename",0,"asmorder.dll",0
db "Copyright",0,"Dwayne Robinson",0
EndVersionData:
%endif
